from sklearn.base import RegressorMixin

from . import model


class XGPRegressor(model.XGPModel, RegressorMixin):
    """ XGP regression model."""

    @property
    def default_loss(self):
        return 'mae'



"""
The goal of this module is to parse an expression returns by XGP and cast it
as a Python function using numpy operators.
"""
import numpy as np


def protected_divide(numerator, denominator):
    where_zeros = denominator != 0
    result = np.divide(numerator, denominator, out=np.zeros_like(numerator), where=where_zeros)
    result[~where_zeros] = 1
    return result


def parse_function(string):
    return {
        # Unary
        'cos': np.cos,
        'sin': np.sin,
        'log': np.log,
        'exp': np.exp,
        'max': np.maximum,
        'min': np.minimum,
        # Binary
        'sum': np.add,
        'sub': np.subtract,
        'div': protected_divide,
        'mul': np.multiply,
        'pow': np.power
    }[string]


def parse_code(code):

    # The code is either a constant or a variable
    if not code.endswith(')'):

        # The code is a variable
        if code.endswith(']'):
            i = int(code[2:len(code)-1])
            return lambda X: X[:, i]

        # The code is a constant
        return lambda X: np.full(shape=len(X), fill_value=float(code), dtype=np.float)

    operator, inside = code[:-1].split('(', 1)

    # Get the appropriate numpy function
    operator = parse_function(operator)

    operands = []
    operand = ''
    parenthesesCounter = 0

    for c in inside:
        if c == ' ':
            continue
        if c == '(':
            parenthesesCounter += 1
        if c == ',' and parenthesesCounter <= 0:
            operands.append(operand)
            operand = ''
        else:
            operand += c
        if c == ')':
            parenthesesCounter -= 1
    operands.append(operand)

    return lambda X: operator(*[parse_code(operand)(X) for operand in operands])



"""
This is a module to be used as a reference for building other modules
"""
import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin, TransformerMixin
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.utils.multiclass import unique_labels
from sklearn.metrics import euclidean_distances


class TemplateEstimator(BaseEstimator):
    """ A template estimator to be used as a reference implementation .

    Parameters
    ----------
    demo_param : str, optional
        A parameter used for demonstation of how to pass and store paramters.
    """
    def __init__(self, demo_param='demo_param'):
        self.demo_param = demo_param

    def fit(self, X, y):
        """A reference implementation of a fitting function

        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_samples, n_features]
            The training input samples.
        y : array-like, shape = [n_samples] or [n_samples, n_outputs]
            The target values (class labels in classification, real numbers in
            regression).

        Returns
        -------
        self : object
            Returns self.
        """
        X, y = check_X_y(X, y)
        # Return the estimator
        return self

    def predict(self, X):
        """ A reference implementation of a predicting function.

        Parameters
        ----------
        X : array-like of shape = [n_samples, n_features]
            The input samples.

        Returns
        -------
        y : array of shape = [n_samples]
            Returns :math:`x^2` where :math:`x` is the first column of `X`.
        """
        X = check_array(X)
        return X[:, 0]**2


class TemplateClassifier(BaseEstimator, ClassifierMixin):
    """ An example classifier which implements a 1-NN algorithm.

    Parameters
    ----------
    demo_param : str, optional
        A parameter used for demonstation of how to pass and store paramters.

    Attributes
    ----------
    X_ : array, shape = [n_samples, n_features]
        The input passed during :meth:`fit`
    y_ : array, shape = [n_samples]
        The labels passed during :meth:`fit`
    """
    def __init__(self, demo_param='demo'):
        self.demo_param = demo_param

    def fit(self, X, y):
        """A reference implementation of a fitting function for a classifier.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
            The training input samples.
        y : array-like, shape = [n_samples]
            The target values. An array of int.

        Returns
        -------
        self : object
            Returns self.
        """
        # Check that X and y have correct shape
        X, y = check_X_y(X, y)
        # Store the classes seen during fit
        self.classes_ = unique_labels(y)

        self.X_ = X
        self.y_ = y
        # Return the classifier
        return self

    def predict(self, X):
        """ A reference implementation of a prediction for a classifier.

        Parameters
        ----------
        X : array-like of shape = [n_samples, n_features]
            The input samples.

        Returns
        -------
        y : array of int of shape = [n_samples]
            The label for each sample is the label of the closest sample
            seen udring fit.
        """
        # Check is fit had been called
        check_is_fitted(self, ['X_', 'y_'])

        # Input validation
        X = check_array(X)

        closest = np.argmin(euclidean_distances(X, self.X_), axis=1)
        return self.y_[closest]


class TemplateTransformer(BaseEstimator, TransformerMixin):
    """ An example transformer that returns the element-wise square root..

    Parameters
    ----------
    demo_param : str, optional
        A parameter used for demonstation of how to pass and store paramters.

    Attributes
    ----------
    input_shape : tuple
        The shape the data passed to :meth:`fit`
    """
    def __init__(self, demo_param='demo'):
        self.demo_param = demo_param

    def fit(self, X, y=None):
        """A reference implementation of a fitting function for a transformer.

        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_samples, n_features]
            The training input samples.
        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : object
            Returns self.
        """
        X = check_array(X)

        self.input_shape_ = X.shape

        # Return the transformer
        return self

    def transform(self, X):
        """ A reference implementation of a transform function.

        Parameters
        ----------
        X : array-like of shape = [n_samples, n_features]
            The input samples.

        Returns
        -------
        X_transformed : array of int of shape = [n_samples, n_features]
            The array containing the element-wise square roots of the values
            in `X`
        """
        # Check is fit had been called
        check_is_fitted(self, ['input_shape_'])

        # Input validation
        X = check_array(X)

        # Check that the input is of the same shape as the one passed
        # during fit.
        if X.shape != self.input_shape_:
            raise ValueError('Shape of input is different from what was seen'
                             'in `fit`')
        return np.sqrt(X)



import abc

from sklearn import base
from sklearn import utils

from . import binding
from . import parse


class XGPModel(abc.ABC, base.BaseEstimator):
    """ Implementation of the scikit-learn API for XGP.

    Parameters
    ----------
    loss_metric : str
        Metric used for scoring program; determines the task to perform.
    funcs : str
        Comma-separated set of authorised functions.
    const_min : float
        Lower bound used for generating constants.
    const_max : float
        Upper bound used for generating constants.
    p_constant : float
        Probability of generating a constant instead of a variable.
    p_full : float
        Probability of using full initialization during ramped half-and-half
        initialization.
    p_terminal : float
        Probability of generating a terminal node during ramped half-and-half
        initialization.
    min_height : int
        Minimum program height used in ramped half-and-half initialization.
    max_height : int
        Minimum program height used in ramped half-and-half initialization.
    n_populations : int
        Number of populations used in the genetic algorithm.
    n_individuals : int
        Number of individuals used for each population in the genetic
        algorithm.
    n_generations : int
        Number of generations used in the GA.
    n_polish_generations : int
        Number of generations used to polish the best program.
    p_hoist_mutation : float
        Probability of applying hoist mutation.
    p_sub_tree_mutation : float
        Probability of applying subtree mutation.
    p_point_mutation : float
        Probability of applying point mutation.
    point_mutation_rate : float
        Probability of modifying an operator during point mutation.
    p_sub_tree_crossover : float
        Probability of applying subtree crossover.
    parsimony_coefficient : float
        Parsimony coefficient by which a program's height is multiplied to
        penalize it's fitness.
    random_state : int, RandomState instance or None, optional (default=None)
        Control the randomization of the algorithm

        - If int, ``random_state`` is the seed used by the random number
          generator;
        - If ``RandomState`` instance, random_state is the random number
          generator;
        - If ``None``, the random number generator is the ``RandomState``
          instance used by ``np.random``.

    Attributes
    ----------
    program_str_ : str
        The string representation of the best program.
    """

    @property
    @abc.abstractmethod
    def default_loss(self):
        pass

    def __init__(self, loss_metric='', funcs='sum,sub,mul,div',
                 const_min=-5, const_max=5, p_constant=0.5, p_full=0.5,
                 p_terminal=0.3, min_height=3, max_height=5, n_populations=1,
                 n_individuals=100, n_generations=30, n_polish_generations=0,
                 p_hoist_mutation=0.1, p_sub_tree_mutation=0.1,
                 p_point_mutation=0.1, point_mutation_rate=0.5,
                 p_sub_tree_crossover=0.3, parsimony_coefficient=0.00001,
                 random_state=None):

        self.loss_metric = loss_metric

        self.funcs = funcs
        self.const_min = const_min
        self.const_max = const_max
        self.p_hoist_mutation = p_hoist_mutation
        self.p_sub_tree_mutation = p_sub_tree_mutation
        self.p_point_mutation = p_point_mutation
        self.point_mutation_rate = point_mutation_rate
        self.min_height = min_height
        self.max_height = max_height

        self.n_populations = n_populations
        self.n_individuals = n_individuals
        self.n_generations = n_generations
        self.n_polish_generations = n_polish_generations
        self.p_constant = p_constant
        self.p_full = p_full
        self.p_terminal = p_terminal
        self.p_sub_tree_crossover = p_sub_tree_crossover

        self.parsimony_coefficient = parsimony_coefficient

        self.random_state = random_state

    def fit(self, X, y, sample_weight=None, eval_set=None, eval_metric=None,
            eval_sample_weight=None, verbose=False):

        X, y = utils.check_X_y(X, y)
        if sample_weight is not None:
            sample_weight = utils.check_array(sample_weight, ensure_2d=False)
        if eval_sample_weight is not None:
            eval_sample_weight = utils.check_array(eval_sample_weight, ensure_2d=False)

        # Extract the validation set
        X_val = None
        y_val = None
        if eval_set is not None:
            X_val = eval_set[0]
            y_val = eval_set[1]
            X_val, y_val = utils.check_X_y(X_val, y_val)

        program_bytes_ = binding.fit(
            X_train=X,
            y_train=y,
            w_train=sample_weight,
            X_val=X_val,
            y_val=y_val,
            w_val=eval_sample_weight,

            loss_metric_name=self.loss_metric if self.loss_metric else self.default_loss,
            eval_metric_name=eval_metric,

            funcs=self.funcs,
            const_min=self.const_min,
            const_max=self.const_max,
            p_constant=self.p_constant,
            p_full=self.p_full,
            p_terminal=self.p_terminal,
            min_height=self.min_height,
            max_height=self.max_height,

            n_populations=self.n_populations,
            n_individuals=self.n_individuals,
            n_generations=self.n_generations,
            n_polish_generations=self.n_polish_generations,
            p_hoist_mutation=self.p_hoist_mutation,
            p_sub_tree_mutation=self.p_sub_tree_mutation,
            p_point_mutation=self.p_point_mutation,
            point_mutation_rate=self.point_mutation_rate,
            p_sub_tree_crossover=self.p_sub_tree_crossover,

            parsimony_coefficient=self.parsimony_coefficient,

            seed=utils.check_random_state(self.random_state).randint(2 ** 24),
            verbose=verbose
        )

        self.program_str_ = program_bytes_.decode('utf-8')

        return self

    def predict(self, X):
        utils.validation.check_is_fitted(self, 'program_str_')
        X = utils.check_array(X)
        return parse.parse_code(self.program_str_)(X)



import ctypes
import glob
import os
import numpy as np


def find_xgp_dll_paths():
    """Finds the XGP dynamic library file."""
    here = os.path.dirname(os.path.abspath(os.path.expanduser(__file__)))

    def make_paths(ext):
        return [
            *glob.glob(os.path.join(here, 'xgp*.{}'.format(ext))),
            *glob.glob(os.path.join(here, '..', 'xgp*.{}'.format(ext))),
            *glob.glob(os.path.join(here, '../..', 'xgp*.{}'.format(ext))),
        ]

    return make_paths('so') + make_paths('dll') + make_paths('pyd')


def load_dll(path):
    """Loads and returns a dynamic library file."""
    return ctypes.cdll.LoadLibrary(path)


class GoString(ctypes.Structure):
    _fields_ = [
        ('p', ctypes.c_char_p),
        ('n', ctypes.c_longlong)
    ]


class GoFloat64Slice(ctypes.Structure):
    _fields_ = [
        ('data', ctypes.POINTER(ctypes.c_double)),
        ('len', ctypes.c_longlong),
        ('cap', ctypes.c_longlong)
    ]


class GoFloat64Matrix(ctypes.Structure):
    _fields_ = [
        ('data', ctypes.POINTER(GoFloat64Slice)),
        ('len', ctypes.c_longlong),
        ('cap', ctypes.c_longlong)
    ]


def numpy_to_slice(arr: np.ndarray) -> GoFloat64Slice:
    # If the slice is 1D then return a GoFloat64Slice
    if len(arr.shape) == 1:
        return GoFloat64Slice(
            arr.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
            len(arr),
            len(arr)
        )
    # If not return a GoFloat64Matrix composed of GoFloat64Slices
    return GoFloat64Matrix(
        (GoFloat64Slice * len(arr))(*[
            numpy_to_slice(arr[i, :])
            for i in range(len(arr))
        ]),
        len(arr),
        len(arr)
    )


def str_to_go_string(s: str):
    return GoString(bytes(s, 'utf-8'), len(s))


def fit(X_train: np.ndarray,
        y_train: np.ndarray,
        w_train: np.ndarray,
        X_val: np.ndarray,
        y_val: np.ndarray,
        w_val: np.ndarray,


        loss_metric_name: str,
        eval_metric_name: str,

        funcs: str,
        const_min: float,
        const_max: float,
        p_constant: float,
        p_full: float,
        p_terminal: float,
        min_height: int,
        max_height: int,

        n_populations: int,
        n_individuals: int,
        n_generations: int,
        n_polish_generations: int,
        p_hoist_mutation: float,
        p_sub_tree_mutation: float,
        p_point_mutation: float,
        point_mutation_rate: float,
        p_sub_tree_crossover: float,

        parsimony_coefficient: float,

        seed: int,
        verbose: bool):
    """Refers to the Fit method in main.go"""

    # Find the and load the XGP dynamic library file
    paths = find_xgp_dll_paths()
    if len(paths) == 0:
        raise RuntimeError("Can't find any XGP dynamic library files")
    lib = load_dll(paths[0])

    # Use Fortran memory layout (and not contiguous memory layout)
    X_train = np.asfortranarray(X_train.astype(float))
    y_train = np.asfortranarray(y_train.astype(float))
    if w_train is not None:
        w_train = np.asfortranarray(w_train.astype(float))
    if X_val is not None:
        X_val = np.asfortranarray(X_val.astype(float))
    if y_val is not None:
        y_val = np.asfortranarray(y_val.astype(float))
    if X_val is not None:
        X_val = np.asfortranarray(X_val.astype(float))

    args = [
        (numpy_to_slice(X_train.T), GoFloat64Matrix),
        (numpy_to_slice(y_train), GoFloat64Slice),
        (numpy_to_slice(w_train if w_train is not None else np.zeros(shape=0)), GoFloat64Slice),
        (numpy_to_slice(X_val.T if X_val is not None else np.zeros(shape=(0, 0))), GoFloat64Matrix),
        (numpy_to_slice(y_val if y_val is not None else np.zeros(shape=0)), GoFloat64Slice),
        (numpy_to_slice(w_val if w_val is not None else np.zeros(shape=0)), GoFloat64Slice),

        (str_to_go_string(loss_metric_name), GoString),
        (str_to_go_string(eval_metric_name if eval_metric_name else ''), GoString),

        (str_to_go_string(funcs), GoString),
        (const_min, ctypes.c_double),
        (const_max, ctypes.c_double),
        (p_constant, ctypes.c_double),
        (p_full, ctypes.c_double),
        (p_terminal, ctypes.c_double),
        (min_height, ctypes.c_longlong),
        (max_height, ctypes.c_longlong),

        (n_populations, ctypes.c_longlong),
        (n_individuals, ctypes.c_longlong),
        (n_generations, ctypes.c_longlong),
        (n_polish_generations, ctypes.c_longlong),
        (p_hoist_mutation, ctypes.c_double),
        (p_sub_tree_mutation, ctypes.c_double),
        (p_point_mutation, ctypes.c_double),
        (point_mutation_rate, ctypes.c_double),
        (p_sub_tree_crossover, ctypes.c_double),

        (parsimony_coefficient, ctypes.c_double),

        (seed, ctypes.c_longlong),
        (verbose, ctypes.c_bool)
    ]

    lib.Fit.argtypes = [arg[1] for arg in args]
    lib.Fit.restype = ctypes.c_char_p

    return lib.Fit(*[arg[0] for arg in args])



import numpy as np

from xgp import binding


def test_numpy_to_slice():
    """Test numpy_to_slice"""
    X = np.array([
        [11, 12, 13],
        [21, 22, 23]
    ])
    go_slice = binding.numpy_to_slice(X)
    assert isinstance(go_slice, binding.GoFloat64Matrix)
    assert go_slice.len == 2
    assert go_slice.cap == 2
    assert go_slice.data[0].len == 3
    assert go_slice.data[0].cap == 3
    assert go_slice.data[1].len == 3
    assert go_slice.data[1].cap == 3



import numpy as np
from sklearn import multiclass
from sklearn.base import ClassifierMixin
from sklearn import utils

from . import model


class XGPClassifier(model.XGPModel, ClassifierMixin):
    """ XGP classification model."""

    @property
    def default_loss(self):
        return 'logloss'

    def fit(self, X, y, **kwargs):
        if len(np.unique(y)) > 2:
            raise ValueError('Multi-class classification is not supported')

        self = super().fit(X, y, **kwargs)
        return self

    def predict(self, X):
        y_pred = super().predict(X)
        return (y_pred > 0.5).astype(int)

    def predict_proba(self, X):
        y_pred = super().predict(X)
        return 1 / (1 + np.exp(-y_pred))



from sklearn import datasets
from sklearn import metrics
from sklearn import model_selection
import xgp


X, y = datasets.load_breast_cancer(return_X_y=True)
X_train, X_test, y_train, y_test = model_selection.train_test_split(X, y)

model = xgp.XGPClassifier(
    loss_metric='f1',
    n_individuals=500,
    n_generations=100,
    random_state=42,
    parsimony_coefficient=0.01
)

model.fit(X_train, y_train, eval_set=(X_test, y_test), verbose=True)

print('Train F1: {:.5f}'.format(metrics.f1_score(y_train, model.predict(X_train))))
print('Test F1: {:.5f}'.format(metrics.f1_score(y_test, model.predict(X_test))))



from sklearn import datasets
from sklearn import metrics
import xgp


X, y = datasets.load_boston(return_X_y=True)

model = xgp.XGPRegressor(
    loss_metric='mae',
    n_individuals=500,
    n_generations=100,
    random_state=42,
    parsimony_coefficient=0.01
)

model.fit(X, y, verbose=True)

print('MAE: {:.5f}'.format(metrics.mean_absolute_error(y, model.predict(X))))



from __future__ import print_function
import sys
from setuptools import Extension, setup, find_packages

with open('requirements.txt') as f:
    INSTALL_REQUIRES = [l.strip() for l in f.readlines() if l]


try:
    import numpy
except ImportError:
    print('numpy is required during installation')
    sys.exit(1)

try:
    import scipy
except ImportError:
    print('scipy is required during installation')
    sys.exit(1)

setup(name='xgp',
      version='0.0.1',
      description='XGP scikit-learn API',
      author='Max Halford',
      packages=find_packages(),
      install_requires=INSTALL_REQUIRES,
      author_email='maxhalford25@gmail.com',
      url='https://maxhalford.github.io/xgp',
      build_golang={'root': 'github.com/MaxHalford/xgp-python'},
      ext_modules=[Extension('xgp', ['xgp/xgp.go'])],
      setup_requires=['setuptools-golang>=0.2.0'],
      )
